module Data.Graph.Clustering.Example where

import Protolude

import Control.Monad (foldM_)
import Data.List (nub, sort)
import Data.Graph.FGL
import Data.Graph.Inductive
import qualified Data.Text as T
import qualified Text.ParserCombinators.Parsec as P
import Text.Parsec.Language (haskellStyle)
import qualified Text.Parsec.Token as PT

import Data.Graph.Clustering.Louvain.Utils
import Data.Graph.Clustering.Louvain.Types
import Data.Graph.Clustering.FLouvain

-- | Run FLouvain.iterate on an example graph
-- Example call:
-- putStrLn $ prettify $ iterateOnce cuiller
-- Prelude.map (fst3 . unCommunity . snd) $ labNodes $ iterateOnce karate
iterateOnce :: Gr a Double -> CGr a
iterateOnce gr = iteration fgr cgr
  where
    fgr = toFGraph    gr
    cgr = initialCGr fgr

runFLouvain :: (Show a, Eq a) => Int -> Int -> FGraph a () -> IO ()
runFLouvain 0 _ fgr = return ()
runFLouvain cycles n fgr = do
  cgr <- runFIterations n fgr
  let fgrNext = louvainSecondStep fgr cgr
  putStrLn ("-----------------" :: Text)
  putStrLn ("New FGraph:" :: Text)
  putStrLn $ prettify fgrNext
  runFLouvain (cycles - 1) n fgrNext

runIterations :: Show a => Int -> Gr a Double -> IO (CGr a)
runIterations n gr = runFIterations n $ toFGraph gr

runFIterations :: Show a => Int -> FGraph a () -> IO (CGr a)
runFIterations n fgr = do
  let fgrWeight = graphWeight fgr
  let initCgr   = initialCGr fgr

  putStrLn ("Initial modularity: " :: Text)
  putStrLn $ T.unpack $ show $ modularity fgr initCgr fgrWeight

  lastCgr <- foldM (runIteration fgr fgrWeight) initCgr [0..n]

  -- at the end, just pretty-print communities
  let coms = filter (not . null . comNodes . llab) (labNodes lastCgr)

  putStrLn ("-------------" :: Text)
  putStrLn ("Non-empty communities: " :: Text)
  mapM_ (\c -> putStrLn (show c :: Text)) coms

  return lastCgr

  where
    runIteration fgr fgrWeight iterCgr i = do
      let iterNextCgr = iteration fgr iterCgr
      putStrLn ("----- ITERATION " <> show i :: Text)
      putStrLn $ prettify iterNextCgr
      putStrLn (show i <> " iteration modularity: " :: Text)
      putStrLn $ T.unpack $ show $ modularity fgr iterNextCgr fgrWeight
      return iterNextCgr

runLouvainFirstStepIterate :: Int -> Gr a Double -> (Modularity, CGr a)
runLouvainFirstStepIterate n gr = (modularity fgr cgr m, cgr)
  where
    fgr = toFGraph gr
    cgr = louvainFirstStepIterate n fgr
    m   = graphWeight fgr

-- | egr <- readPythonGraph "<file-path>"
-- let gr = head $ Data.Either.rights [egr]
readPythonGraph :: FilePath -> IO (Either P.ParseError (Gr () Double))
readPythonGraph src = do
  contents <- readFile src
  let eEdges = P.parse edgeParser "(unknown)" $ T.unpack contents
  case eEdges of
    Left err -> do
      return $ Left err
    Right edges -> do
      return $ Right $ mkGraph' edges
  where
    lexer = PT.makeTokenParser haskellStyle
    edgeParser :: P.GenParser Char st [(Node, Node, Double)]
    edgeParser = P.many edgeLine
    edgeLine :: P.GenParser Char st (Node, Node, Double)
    edgeLine = do
      source <- edgeSource
      target <- edgeTarget
      weight <- edgeWeight
      return (source, target, weight)
    edgeSource :: P.GenParser Char st Node
    edgeSource = do
      v <- PT.integer lexer
      return $ fromIntegral v
    edgeTarget :: P.GenParser Char st Node
    edgeTarget = do
      v <- PT.integer lexer
      return $ fromIntegral v
    edgeWeight :: P.GenParser Char st Double
    edgeWeight = do
      v <- PT.float lexer
      return v

karate :: Gr () Double
-- karate =  mkGraph' <$> importGraphFromGexf "src/Data/karate.gexf"
karate = mkGraph [(1,()),(2,()),(3,()),(4,()),(5,()),(6,()),(7,()),(8,()),(9,()),(10,()),(11,()),(12,()),(13,()),(14,()),(15,()),(16,()),(17,()),(18,()),(19,()),(20,()),(21,()),(22,()),(23,()),(24,()),(25,()),(26,()),(27,()),(28,()),(29,()),(30,()),(31,()),(32,()),(33,()),(34,())] [(1,2,1.0),(1,3,1.0),(1,4,1.0),(1,5,1.0),(1,6,1.0),(1,7,1.0),(1,8,1.0),(1,9,1.0),(1,11,1.0),(1,12,1.0),(1,13,1.0),(1,14,1.0),(1,18,1.0),(1,20,1.0),(1,22,1.0),(1,32,1.0),(2,3,1.0),(2,4,1.0),(2,8,1.0),(2,14,1.0),(2,18,1.0),(2,20,1.0),(2,22,1.0),(2,31,1.0),(3,4,1.0),(3,8,1.0),(3,9,1.0),(3,10,1.0),(3,14,1.0),(3,28,1.0),(3,29,1.0),(3,33,1.0),(4,8,1.0),(4,13,1.0),(4,14,1.0),(5,7,1.0),(5,11,1.0),(6,7,1.0),(6,11,1.0),(6,17,1.0),(7,17,1.0),(9,31,1.0),(9,33,1.0),(9,34,1.0),(10,34,1.0),(14,34,1.0),(15,33,1.0),(15,34,1.0),(16,33,1.0),(16,34,1.0),(19,33,1.0),(19,34,1.0),(20,34,1.0),(21,33,1.0),(21,34,1.0),(23,33,1.0),(23,34,1.0),(24,26,1.0),(24,28,1.0),(24,30,1.0),(24,33,1.0),(24,34,1.0),(25,26,1.0),(25,28,1.0),(25,32,1.0),(26,32,1.0),(27,30,1.0),(27,34,1.0),(28,34,1.0),(29,32,1.0),(29,34,1.0),(30,33,1.0),(30,34,1.0),(31,33,1.0),(31,34,1.0),(32,33,1.0),(32,34,1.0),(33,34,1.0)]


karate2com :: [[Node]]
karate2com = sort $ map (sort) [[10, 29, 32, 25, 28, 26, 24, 30, 27, 34, 31, 33, 23, 15, 16, 21, 19], [3, 9, 8, 4, 14, 20, 2, 13, 22, 1, 18, 12, 5, 7, 6, 17]]


eU :: [LEdge Double]
eU = [
     (2,1,1)
    ,(1,2,1)
    
    ,(1,4,1)
    ,(4,1,1)
    
    ,(2,3,1)
    ,(3,2,1)
    
    ,(3,4,1)
    ,(4,3,1)
    
    ,(4,5,1)
    ,(5,4,1)
--    ,(6,7,1)
    ]

eD :: [LEdge Double]
eD = [
     (2,1,1)
    
    ,(1,4,1)
    
    ,(2,3,1)
    
    ,(3,4,1)
    
    ,(4,5,1)
    ]

gU :: Gr () Double
gU = mkGraph' eU

-- > prettyPrint gU
-- 1:()->[(1,2),(1,4)]
-- 2:()->[(1,1),(1,3)]
-- 3:()->[(1,2),(1,4)]
-- 4:()->[(1,1),(1,3),(1,5)]
-- 5:()->[(1,4)]

cuiller :: Gr () Double
cuiller = gU

-- Visual representation:
--    2
--   / \
--  1   3
--   \ /
--    4
--    |
--    5
-- 
-- 

gD :: Gr () Double
gD = mkGraph' eD

eD' :: [LEdge Double]
eD' = [
     (2,1,1)
    
    ,(1,4,1)
    
    ,(2,3,1)
    
    ,(3,4,1)
    
    ,(4,5,1)
    ,(5,6,1)
    ,(5,7,1)
    ,(6,7,1)
    ]
gD' :: Gr () Double
gD' = mkGraph' eD'
