{-|
Module      : Data.Graph.Clustering.FLouvain
Description : Purely functional (Inductive) Louvain clustering
Copyright   : (c) Alexandre Delanoë, CNRS, 2020-Present
License     : AGPL + CECILL v3
Maintainer  : alexandre.delanoe+louvain@iscpif.fr
Stability   : experimental
Portability : POSIX

# Louvain Algorithm

Our algorithm is divided into two phases that are repeated iteratively.
Assume that we start with a weighted network of N nodes. First, we
assign a different community to each node of the network. So, in this
initial partition there are as many communities as there are nodes.
Then, for each node i we consider the neighbours j of i and we evaluate
the gain of modularity that would take place by removing i from its
community and by placing it in the community of j. The node i is then
placed in the community for which this gain is maximum (in the case of
a tie we use a breaking rule), but only if this gain is positive. If
no positive gain is possible, i stays in its original community. This
process is applied repeatedly and sequentially for all nodes until
no further improvement can be achieved and the first phase is then
complete.

Reference:
Blondel, Vincent D; Guillaume, Jean-Loup; Lambiotte, Renaud;
Lefebvre, Etienne (9 October 2008). "Fast unfolding of communities
in large networks". Journal of Statistical Mechanics: Theory and
Experiment. 2008 (10): P10008. arXiv:0803.0476 Freely accessible.
doi:10.1088/1742-5468/2008/10/P10008.

-}

module Data.Graph.Clustering.FLouvain
  where

import Protolude
import Data.Graph.Inductive
import qualified Data.List as DL

import Data.Graph.FGL
import Data.Graph.Clustering.Louvain.Utils (fixPt, mkFGraph)
import Data.Graph.Clustering.Louvain.Types

------------------------------------------------------------------------

-- | Main Louvain first step iteration function
louvainFirstStepIterate :: Int -> FGraph a b -> CGr a
louvainFirstStepIterate n gr = fixPt n iterator cond initCGr
  where
    initCGr      = initialCGr  gr
    grWeight     = graphWeight gr
    iterator cgr = iteration   gr cgr
    cond cgr     = (unModularity $ modularity gr cgr grWeight) < 0.1

-- | Second step from the Louvain paper -- given a clustering, create new graph
-- of clusters
louvainSecondStep :: forall a b c. Eq c => FGraph a b -> CGr c -> FGraph (Community c) ()
louvainSecondStep gr cgr = mkFGraph nodes edges
  where
    nodes :: [LNode (Community c)]
    --nodes = filter (\(_, com) -> length (comNodes com) > 0) $ labNodes cgr
    nodes = labNodes cgr
    edges :: [(Node, Node, Double)]
    edges = concatMap comEdges $ labNodes cgr
    comEdges :: LNode (Community c) -> [(Node, Node, Double)]
    comEdges lnCom = mapMaybe (comToComEdge lnCom) $ labNodes cgr
    comToComEdge :: LNode (Community c) -> LNode (Community c) -> Maybe (Node, Node, Double)
    -- No self-edges
    comToComEdge lnCom1 lnCom2 | lnCom1 == lnCom2            = Nothing
    comToComEdge (_, com1) _   | length (comNodes com1) == 0 = Nothing
    comToComEdge _ (_, com2)   | length (comNodes com2) == 0 = Nothing
    comToComEdge (com1N, com1) (com2N, com2)      = Just (com1N, com2N, comToComWeight gr com1 com2)

-- | Weight between communities. Base graph is needed to fetch weights between
-- individual nodes.
comToComWeight :: FGraph a b -> Community c -> Community c -> Double
comToComWeight gr com1 com2 = weight
  where
    weight :: Double
    weight = sum $ map (\n -> unNodeComWeightSum $ nodeComWeightSum com2 $ context gr n) $ comNodes com1

------------------------------------------------------------------------


-- | Compute initial 'CGr' for a given 'FGraph a b'. This means, put each node
-- in a separate community.
initialCGr :: FGraph a b -> CGr a
initialCGr gr = gmap singletonCom gr
  where
    -- A singleton community is given:
    -- the same node id for a community
    -- same incoming/outgoing edges
    -- custom Community label
    singletonCom (p, v, l, s) = (p', v, Community ([v], iws, tws, l), s')
      where
        p' = map edgeComRemap p
        s' = map edgeComRemap s
        edgeComRemap (_, n) = ((), n)
        edges = lneighbors gr v
        -- no internal links
        iws = InWeightSum 0.0
        -- Just sum over the edges coming into/out of node v.  This is because
        -- there are no inner links, all other nodes are external.
        tws = TotWeightSum $ sumEdgeWeights edges

-- ALGORITHM

-- | Q function from Louvain paper (1).
-- We just fold over the communities (this is because of the delta(c_i, c_j)
-- param)
modularity :: FGraph a b -> CGr c -> GraphWeightSum -> Modularity
modularity gr cgr m = Modularity $ coeff * ( ufold modularity' 0.0 cgr )
  where
    coeff = 0.5 / (unGraphWeightSum m)
    -- Sum over nodes in community
    -- \Sum A_ij is just the InWeightSum (in formula (1) it is sum of weights
    -- between nodes i and j both in the same community, as enforced by the
    -- \delta(c_i, c_j) symbol)
    modularity' (_, _, com, _) acc = acc + component
      where
        component = (unInWeightSum $ comInWeightSum com) - weightsMul
        weightsMul = coeff * ( sum $ map weightsMul' $ comNodes com )
        weightsMul' n = (ki n) * (sum $ map ki $ comNodes com)
        -- k_i variable in formula (1)
        ki :: Node -> Double
        ki n = unNodeWeightSum $ nodeWeightSum $ context gr n

type Delta c = Community c -> NodeWeightSum -> NodeComWeightSum -> GraphWeightSum -> DeltaQ
-- | Delta Q function from Louvain paper (2).
delta :: Delta c
delta com ki kiin m = DeltaQ $ acc - dec
  where
    inWeightSum = comInWeightSum com
    totWeightSum = comTotWeightSum com
    acc = accL - accR * accR
    accL = 0.5 * (unInWeightSum inWeightSum + 2.0 * (unNodeComWeightSum kiin)) / (unGraphWeightSum m)
    accR = 0.5 * (unTotWeightSum totWeightSum + unNodeWeightSum ki) / (unGraphWeightSum m)
    dec = decL - decM * decM - decR * decR
    decL = 0.5 * (unInWeightSum inWeightSum) / (unGraphWeightSum m)
    decM = 0.5 * (unTotWeightSum totWeightSum) / (unGraphWeightSum m)
    decR = 0.5 * (unNodeWeightSum ki) / (unGraphWeightSum m)

-- | One iteration step takes the graph and existing communities as a graph and
-- computes new community graph

-- NOTE: xdfsFoldWith only iterates with the defined Context. In each step of
-- the algorithm, we need full node information however (i.e. all edges
-- connected to a node), so instead of calling just 'step' we are forced to call via:
-- 'step . context gr . node''
-- We could avoid the higher complexity, eg. by precomputing the whole graph
-- into a HashMap Node [Edge].
iteration :: FGraph a b -> CGr c -> CGr c
iteration gr cs = xdfsFoldWith suc' (\(_, v, _, _)
                -> step gw $ context gr $ v) cs (nodes gr) gr
  where
    gw = graphWeight gr
    --weightSum = ufold weightSum' 0 gr
    --weightSum' (p, v, l, s) acc = acc + (sum )

-- TODO Remember to filter out empty Communities
-- | Step for one node. We try re-assign it to a neighbouring community, where
-- the increase of modularity for graph will be the largest
step :: forall a b c. GraphWeightSum -> CFunFold a (FEdge b) (CGr c)
step gw ctx@(_, v, _, _) cgr = newCgr
  where
    newCgr = case mNc of
      Nothing -> cgr
      Just nc ->
        if bestFitdq > 0.0 then
          let newBestFitCom = makeMove Into bestFitCom
              newNc = makeMove OutOf nc
              in
            replaceLNode (replaceLNode cgr newNc) newBestFitCom
        else
          cgr

    (bestFitCom, DeltaQ bestFitdq) =
      maximumBy (\(_, deltaq1) (_, deltaq2) -> compare deltaq1 deltaq2) deltas

    mNc :: Maybe (LNode (Community c))
    mNc = nodeCommunity v cgr
    ncs :: [LNode (Community c)]
    ncs = nodeNeighbours v cgr

    -- We move node from community nc into ncs
    -- moves :: Maybe (LNode Community, [LNode Community])
    -- moves = case mNc of
    --   Nothing -> Nothing
    --   Just nc -> Just ( makeMove OutOf nc
    --                   , map (makeMove Into) ncs )

    makeMove :: Direction -> LNode (Community c) -> LNode (Community c)
    makeMove direction (cn, c) = (cn, moveNodeWithContext ctx direction c)

    -- k_i variable in formula (2)
    ki :: NodeWeightSum
    ki = nodeWeightSum ctx

    deltas :: [(LNode (Community c), DeltaQ)]
    deltas = map (\c -> (c, delta' c)) ncs

    delta' :: LNode (Community c) -> DeltaQ
    delta' com = delta (llab com) ki kiin gw
      where
        -- k_i,in variable in formula (2)
        kiin :: NodeComWeightSum
        kiin = nodeComWeightSum (llab com) ctx

    -- TODO Compute \Delta Q (gain of moving node v into Community C) which consists of:
    -- - Community InWeightSum
    -- - Community TotWeightSum
    -- - sum of weights of links incident to node v (taken from 'p' and 's')
    -- - sum of weights of links from node v to nodes in C (taken from 'p' and 's')
    -- - sum of weights of all the links in the network
    -- So the Delta function takes as parameters:
    -- C, (edges from C in cgr), (edges from v in gr), (edges from v to C), (edges in gr)

-- COMMUNITY GRAPH FUNCTIONS

-- | 'Direction' when moving node 'Into'/'OutOf' community
data Direction = Into | OutOf

-- | Given 'Node' and 'Community' graph, find the 'LNode' of 'Community' which
-- contains the node
nodeCommunity :: Node -> CGr c -> Maybe (LNode (Community c))
nodeCommunity n cgr = head (filter f $ labNodes cgr)
  where
    f :: (a, Community c) -> Bool
    f (_, com) = n `elem` comNodes com

-- | Find 'LNode's of 'Community' graph neighbouring a given node
nodeNeighbours :: Node -> CGr c -> [LNode (Community c)]
nodeNeighbours n cgr =
  case nodeCommunity n cgr of
    Nothing        -> []
    Just (cn, _)  -> mapMaybe (lnode cgr) (neighbors cgr cn)

-- | Find 'Ajd CGrEdge's of 'Community' graph neighbouring a given node
-- nodeLNeighbours :: Node -> CGr -> Adj CGrEdge
-- nodeLNeighbours n cgr =
--   case nodeCommunity n cgr of
--     Nothing        -> []
--     Just (cn, _) -> lneighbors cgr cn

-- | Moves 'Node' in the 'Direction' of 'Community' and recomputes 'Community''s weights
moveNode :: forall a b c. FGraph a b -> Node -> Direction -> Community c -> Community c
moveNode gr n direction c = moveNodeWithContext ctx direction c
  where
    ctx :: Context a (FEdge b)
    ctx = context gr n

-- | Same as 'moveNode' above but with only node context, not whole graph
moveNodeWithContext :: forall a b c. Context a (FEdge b) -> Direction -> Community c -> Community c
moveNodeWithContext ctx@(_, n, _, _) direction com@(Community (ns, inwsum, totwsum, l)) =
  Community (newNs, InWeightSum newInWsum, TotWeightSum newTotWsum, l)
    where
      newNs = case direction of
        Into  -> sort (n:ns)
        OutOf -> DL.delete n ns

      (newInWsum, newTotWsum) =
        case direction of
          Into  -> (unInWeightSum inwsum + unNodeComWeightSum cws, unTotWeightSum totwsum - unNodeComWeightSum cws + unNodeNonComWeightSum nws)
          OutOf -> (unInWeightSum inwsum - unNodeComWeightSum cws, unTotWeightSum totwsum + unNodeComWeightSum cws - unNodeNonComWeightSum nws)

      -- sum of edge weights inside community
      cws :: NodeComWeightSum
      cws = nodeComWeightSum com ctx

      nws :: NodeNonComWeightSum
      nws = nodeNonComWeightSum com ctx


{-
-- | Moves 'Node' between two 'Community'ies, recomputing their weights
moveNode' :: FGraph a -> Node -> (Community, Community) -> (Community, Community)
moveNode' gr n cFromTo = moveNodeWithNeighbours' lnNeighbors n cFromTo
  where
    lnNeighbors :: Adj FEdge
    lnNeighbors = lneighbors gr n

-- | Same as 'moveNode' but with direct neighbours list
moveNodeWithNeighbours' :: Adj FEdge -> Node -> (Community, Community) -> (Community, Community)
moveNodeWithNeighbours' lnNeighbors n ((Community (fromNs, fromInWSum, fromTotWSum)), (Community (toNs, toInWSum, toTotWSum))) =
  (newFrom, newTo)
  where
    newFrom = Community (DL.delete n fromNs, newFromInWSum, newFromTotWSum)
    newTo = Community (n:toNs, newToInWSum, newToTotWSum)

    -- Node is removed, so we reduce the internal weight sum
    newFromInWSum = fromInWSum - sumFromN
    newFromTotWSum = fromTotWSum + sumFromN - sumFromNonN

    -- Node is added, so we increase the internal weight sum
    newToInWSum = toInWSum + sumToN
    newToTotWSum = toTotWSum + sumToN - sumToNonN

    sumFromN :: InWeightSum
    sumFromN = sum $ map fst fromComNeighbors
    sumFromNonN :: TotWeightSum
    sumFromNonN = sum $ map fst fromNonComNeighbors

    sumToN :: InWeightSum
    sumToN = sum $ map fst toComNeighbors
    sumToNonN :: TotWeightSum
    sumToNonN = sum $ map fst toNonComNeighbors

    fromComNeighbors :: Adj FEdge
    fromComNeighbors = filter (\ln -> snd ln `elem` fromNs) lnNeighbors
    fromNonComNeighbors :: Adj FEdge
    fromNonComNeighbors = filter (\ln -> snd ln `notElem` fromNs) lnNeighbors
    toComNeighbors :: Adj FEdge
    toComNeighbors = filter (\ln -> snd ln `elem` toNs) lnNeighbors
    toNonComNeighbors :: Adj FEdge
    toNonComNeighbors = filter (\ln -> snd ln `notElem` toNs) lnNeighbors
-}
