{-|
Module      : Data.Graph.Clustering.ILouvain
Description : Purely functional (Inductive) Louvain clustering
Copyright   : (c) Alexandre Delanoë, CNRS, 2020-Present
License     : AGPL + CECILL v3
Maintainer  : alexandre.delanoe+louvain@iscpif.fr
Stability   : experimental
Portability : POSIX

ILouvain: really inductive Graph

-}

module Data.Graph.Clustering.ILouvain
  where

import Data.Set (fromList)
import Data.Maybe (catMaybes)
import Data.List (zip, cycle)
import Protolude hiding (empty, (&))
import Data.Graph.Inductive
import qualified Data.Graph.Clustering.HLouvain as H

------------------------------------------------------------------------
-- HyperGraph Definition
type HyperGraph   a b = Gr (Gr () a) b
type HyperContext a b = Context (Gr () a) b

-- TODO Later (hypothesis still)
-- type StreamGraph  a b = Gr a (Gr () b)


hnodes :: HyperGraph a b -> Node -> [Node]
hnodes g n = case match n g of
  (Nothing, _) -> []
  (Just (p, n, l, s), _) -> n : nodes l

hedges :: HyperGraph a b -> Node -> [Edge]
hedges = undefined

hneighbors :: HyperGraph a b -> Node -> [Node]
hneighbors = undefined

------------------------------------------------------------------------
modularity :: HyperGraph a b -> [Node] -> Double
modularity g ns = H.modularity g (fromList ns)

------------------------------------------------------------------------
-- Spoon Graph
--    1
--   / \
--  2   3
--   \ /
--    4
--    |
--    5
spoon :: HyperGraph Double Double
spoon = mkGraph ns es
  where
    ns :: [LNode (Gr () Double)]
    ns = zip [1..6] (cycle [empty])

    es :: [LEdge Double]
    es = [ (1, 2, 1.0)
         , (1, 3, 1.0)
         , (2, 4, 1.0)
         , (3, 4, 1.0)
         , (4, 5, 1.0)
         ]

-- | Needed functions (WIP)
-- mv: a Node elsewhere in the HyperGraph

-- Move Properties:
-- mv Node into another Node and remove it again should be equal to
-- identity

-- property test:
-- mv (mv g [a] [b]) [b,a] [] = identity
-- mv g [a] [] == g
-- mv g [a,b] [] /= g
-- mv (mv spoon [1] [4]) [4,1] [] = identity

-- Move target type
-- let's start simple: path lenght <= 2 max
mv :: HyperGraph a a
   -> [Node] -> [Node]
   -> HyperGraph a a
mv g [ ] [ ] = g
mv g [_] [ ] = g
mv g [ ] [_] = g

mv g [a] [b] = case a == b of
  True  -> panic "mv, impossible: moved node is same as destination"
  False -> mv' g a b
mv g [a,b] [ ] = case match a g of
  (Nothing,  _) -> panic "mv: fst Node of Path does not exist"
  (Just (p, n, l, s), g1) -> case match b l of
    (Just (p',n',l',s'), g2)  -> (p', n', g2, s')
                               & ((p , n , delNode b l , s )
                               & g1)
    (Nothing, _) -> panic "mv: snd Node of Path does not exist"

mv g (x:xs) (y:ys) = panic "mv: path too long"

                  ----------------------------

-- | Start simple (without path)
mv' :: HyperGraph a a
    -> Node -> Node
    -> HyperGraph a a
mv' g n1 n2 =  (mvMContext c1 c2) & g2
  where
    (c1, g1) = match n1 g
    (c2, g2) = match n2 g1

mvMContext :: Maybe (HyperContext a a)
           -> Maybe (HyperContext a a)
           -> HyperContext a a
mvMContext (Just (a1 ,n ,l ,a2 ))
           (Just (a1',n',l',a2')) = (a1',n', merge l' (c & l),a2')
            where
              c = (a1, n, (), a2)

mvMContext _ (Just _) = panic "mvmcontext: First Node  does not exist"
mvMContext (Just _) _ = panic "mvmcontext: Snd   Node  does not exist"
mvMContext _ _        = panic "mvmcontext: Both  Nodes do   not exist"

merge :: (Graph gr, DynGraph gr)
      => gr a b -> gr a b -> gr a b
merge = ufold (&)

------------------------------------------------------------------------
test_mv :: Ord a => HyperGraph a a -> Node -> Node -> Bool
test_mv g a b = (mv (mv g [a] [b]) [b,a] []) == g
------------------------------------------------------------------------
-- | Recursive Node of Graph
{-
rnodes :: RGraph -> [Node]
rnodes Empty = []
rnodes g     = concat $ map (\(x1, x2) -> [x1] <> rnodes x2) $ labNodes g

rlabNodes :: Graph' a b -> [LNode a]
rlabNodes Empty' = []
rlabNodes g      = labNodes g
-}

------------------------------------------------------------------------
-- Paths in the Graph to be tested
-- Directed graph strategy
path_dir :: Graph gr => gr a b -> [[Node]]
path_dir g = map (\xs -> dfs xs g ) (components g)

-- UnDirected graph strategy
path :: (DynGraph gr, Eq b) => gr a b -> [[Node]]
path g' = map sortNodes cs
  where
    sortNodes ns = case head $ sortOn (Down . (deg g)) ns of
      Nothing -> []
      Just  n -> dfs [n] g  -- dfs for glustering, bfs for klustering
    cs = components g
    g  = undir g'
------------------------------------------------------------------------
------------------------------------------------------------------------
