{-|
Module      : Data.Louvain.Formats.Gexf
Description : 
Copyright   : (c) CNRS, 2017-Present
License     : AGPL + CECILL v3
Maintainer  : alexandre.delanoe+louvain@iscpif.fr
Stability   : experimental
Portability : POSIX

Tools to clustering Graphs with louvain clustering algorithm (not optimized version).

References:
 * Blondel, Vincent D; Guillaume, Jean-Loup; Lambiotte, Renaud; Lefebvre, Etienne (9 October 2008). "Fast unfolding of communities in large networks". Journal of Statistical Mechanics: Theory and Experiment. 2008 (10): P10008. arXiv:0803.0476 Freely accessible. doi:10.1088/1742-5468/2008/10/P10008.
 * Louvain_Modularity https://en.wikipedia.org/wiki/Louvain_Modularity
-}

{-# LANGUAGE ImplicitPrelude #-}

module Data.Graph.Clustering.Louvain 
  where

import Data.List (maximumBy, nub, intersect, foldl', zipWith, concat)
import Data.Graph.Inductive
import Data.Graph.Clustering.Louvain.Utils (LouvainNode(..), toFGraph)
import Data.Graph.Clustering.FLouvain (louvainFirstStepIterate, initialCGr)
import Data.Graph.Clustering.Louvain.Types (Community(..), comNodes)
------------------------------------------------------------------------
-- | Definitions
------------------------------------------------------------------------
type Modularity = Double
type Community = [Node]
-- type Community' = Community { nodes :: [Node], modularity :: Maybe Modularity}
-- type Partition = [Community]
type Reverse = Bool

------------------------------------------------------------------------
flouvain :: Int -> Gr () Double -> [[Node]]
flouvain n g = map (comNodes . snd) $ labNodes g'
  where
    g' = louvainFirstStepIterate n (toFGraph g)
------------------------------------------------------------------------
hLouvain :: (Eq b, DynGraph gr)
         => Reverse
         -> gr a b
         -> [LouvainNode]
hLouvain r g = concat $ toLouvainNode (bestpartition r g)
  where
    toLouvainNode :: [[Node]] -> [[LouvainNode]]
    toLouvainNode ns = zipWith (\cId ns' -> map (\n -> LouvainNode n cId) ns')
                               [1..] ns
------------------------------------------------------------------------
-- | Partitionning the graph
------------------------------------------------------------------------
bestpartition :: (Eq b, DynGraph gr)
              => Reverse
              -> gr a b
              -> [[Node]]
bestpartition r gr = converge gr (start gr r)

converge :: (Eq b, DynGraph gr)
         => gr a1 b
         -> [[Node]]
         -> [[Node]]
converge gr ns = case stepscom gr (length ns) ns of
            ns' | ns == ns' -> ns
                | otherwise -> stepscom gr (length ns') ns'

------------------------------------------------------------------------
dendogram :: (Eq b, DynGraph gr) => gr a b -> Int -> Reverse -> [[Node]]
dendogram gr n r = stepscom gr n (start gr r)

start :: DynGraph gr => gr a b -> Reverse -> [[Node]]
start gr r = order' $ Prelude.map (\x -> [x]) ( nodes gr )
    where
        order' = case r of
                  True  -> reverse
                  False -> id

------------------------------------------------------------------------
------------------------------------------------------------------------
stepscom :: (DynGraph gr, Eq b) => gr a1 b -> Int -> [[Node]] -> [[Node]]
stepscom gr n ns = foldl' (\xs _ -> stepcom gr' (smallCom xs) xs) ns [1..n]
    where
        gr' = undir gr
        smallCom xs = head $ filter (\x -> length x == minimum (Prelude.map length xs))
                                    (reverse xs)

------------------------------------------------------------------------
stepcom :: DynGraph gr => gr a b -> [Node] -> [[Node]] -> [[Node]]
stepcom gr n ns = bestModularity gr
                $ [ns] ++ Prelude.map (\x -> x ++ neighout) (addcom n neighin)
    where
        -- | First remove the node (n) of the current partition (ns)
        ns' = filter (/= n) ns
        neighin    = filter (\c ->  (intersect (neighcom gr n) c) /= [] ) ns'
        neighout   = filter (\c ->  (intersect (neighcom gr n) c) == [] ) ns'

addcom :: [a] -> [[a]] -> [[[a]]]
addcom com coms = Prelude.map (\n -> addcom' com (rotate n coms)) ns
    where
        ns = [0.. fromIntegral (length coms) ]
        addcom' c cs = [com'] ++ coms''
            where
                com'   = concat $ [c] ++ (take 1 cs)
                coms'' = drop 1 cs

neighcom :: DynGraph gr => gr a b -> [Node] -> [Node]
neighcom gr ns = ( nub . filter (not . (`elem` ns)) . concat ) ns'
    where ns' = Prelude.map (neighbors gr) ns


rotate :: Int -> [a] -> [a]
rotate _ [] = []
rotate n xs = zipWith const (drop n (cycle xs)) xs

------------------------------------------------------------------------
-- | Computing modularity of the partition
------------------------------------------------------------------------
bestModularity :: DynGraph gr
                 => gr a b
                 -> [[[Node]]]
                 ->  [[Node]]
bestModularity gr ns = maximumBy (compareModularities gr) ns

compareModularities :: DynGraph gr
                    => gr a b
                    -> [[Node]]
                    -> [[Node]]
                    -> Ordering
compareModularities gr xs ys
    | modularities gr xs < modularities gr ys = LT
    | modularities gr xs > modularities gr ys = GT
    | otherwise = EQ

modularities :: DynGraph gr => gr a b -> [[Node]] -> Double
modularities gr xs = sum $ Prelude.map (\y -> modularity gr y) xs

modularity :: DynGraph gr => gr a b -> [Node] -> Double
modularity gr ns = coverage - edgeDensity
    where
        coverage :: Double
        coverage  = sizeSubGraph / sizeAllGraph 
            where
                sizeSubGraph :: Double
                sizeSubGraph = fromIntegral ( size $ subgraph ns gr )

                sizeAllGraph :: Double
                sizeAllGraph = fromIntegral (size gr)

        edgeDensity :: Double
        edgeDensity = (sum (Prelude.map (\node -> (degree node) / links ) ns)) ** 2
            where
                degree :: Node -> Double
                degree node = fromIntegral (deg gr node)

                links :: Double
                links       = fromIntegral (2 * (size gr))

----------------------------------------------------------
-- | Exampel to discover what NP complete means
----------------------------------------------------------
bestPartition' :: DynGraph gr => gr a b -> [[Node]]
bestPartition' gr = maximumBy (compareModularities gr) $ gpartition gr

gpartition :: DynGraph gr => gr a b -> [[[Node]]]
gpartition gr = separate (nodes gr)

takeDrop :: Int -> [a] -> [[a]]
takeDrop n xs = [ (take n xs), drop n xs]

-- | Naive description of the algo
-- separate' :: forall a. [a] -> [[[a]]]
-- separate' xs = [ takeDrop t (rotate r xs)
--                | t <- [1.. fromIntegral (length xs) - 1 ]
--                , r <- [0.. fromIntegral (length xs)     ]
--                ]
-- Optimization
-- http://stackoverflow.com/questions/35388734/list-partitioning-implemented-recursively
separate :: [a] -> [[[a]]]
separate [] = [[]]
separate (x:xs) = let recur = separate xs
                      split = do
                        partition <- recur
                        pure $ [x] : partition
                      noSplit = do
                        (y:ys) <- recur
                        pure $ (x:y):ys
                  in split ++ noSplit

----------------------------------------------------------
