module Data.Graph.Clustering.Louvain.Types where

import Protolude

import Data.Graph.Inductive


--  "glue" : function to gather/merge communities
--  "klue" : function to split communities
data ClusteringMethod = Glue | Klue
  deriving (Eq)

newtype Weight = Weight { unWeight :: Double }
  deriving (Show, Eq, Ord)
type FEdge b = (Weight, b)
fedgeWeight :: FEdge b -> Double
fedgeWeight = unWeight . fst
sumEdgeWeights :: Adj (FEdge b) -> Double
sumEdgeWeights es = sum $ map (fedgeWeight . fst) es

-- Our basic graph. Nodes have custom labels. Edges have weight assigned to them.
type FGraph a b = Gr a (FEdge b)

-- | Used for k_i in formula (2)
-- (sum of the weights of the links incident to node i)
newtype NodeWeightSum = NodeWeightSum { unNodeWeightSum :: Double }
  deriving (Show, Eq, Ord)
nodeWeightSum :: Context a (FEdge b) -> NodeWeightSum
nodeWeightSum (p, _, _, s) = NodeWeightSum $ sumEdgeWeights $ p <> s

-- | Used for k_i,in in formula (2)
-- (Sum of weights of links from a given 'Node' to nodes in a given 'Community')
newtype NodeComWeightSum = NodeComWeightSum { unNodeComWeightSum :: Double }
  deriving (Show, Eq, Ord)
nodeComWeightSum :: Community c -> Context a (FEdge b) -> NodeComWeightSum
nodeComWeightSum com (p, _, _, s) =
  NodeComWeightSum $ sumEdgeWeights $ filter (\(_, n) -> n `elem` comNodes com) $ p <> s

newtype NodeNonComWeightSum = NodeNonComWeightSum { unNodeNonComWeightSum :: Double }
  deriving (Show, Eq, Ord)
nodeNonComWeightSum :: Community c -> Context a (FEdge b) -> NodeNonComWeightSum
nodeNonComWeightSum com (p, _, _, s) =
  NodeNonComWeightSum $ sumEdgeWeights $ filter (\(_, n) -> n `notElem` comNodes com) $ p <> s

-- Probably this structure is better to reduce the number of computations
-- (precompute sum of node weights, which is the k_i variable in formula (2)).
-- type FNode a = (NodeWeightSum, a)
-- fnodeWeightSum :: FNode a -> Double
-- fnodeWeightSum = unNodeWeightSum . fst

-- | This is the m variable in formula (2) of the Louvain paper
newtype GraphWeightSum = GraphWeightSum { unGraphWeightSum :: Double }
  deriving (Show, Eq, Ord)
-- | This is the \Sum_in in formula (2) of the Louvain paper
-- (sum of the weights of the links inside C)
newtype InWeightSum = InWeightSum { unInWeightSum :: Double }
  deriving (Show, Eq, Ord)
-- | This is the \Sum_tot in formula (2) of the Louvain paper
-- (sum of the weights of the links incident to nodes in C)
newtype TotWeightSum = TotWeightSum { unTotWeightSum :: Double }
  deriving (Show, Eq, Ord)
-- | Computed Delta_Q value in (2)
newtype DeltaQ = DeltaQ { unDeltaQ :: Double }
  deriving (Show, Eq, Ord)
-- | Computed modularity in (1)
newtype Modularity = Modularity { unModularity :: Double }
  deriving (Show, Eq, Ord)

-- | Type for the clusters we will be creating.
newtype Community a = Community { unCommunity ::  ([Node], InWeightSum, TotWeightSum, a) }
  deriving (Show, Eq, Ord)
comNodes :: Community c -> [Node]
comNodes (Community (ns, _, _, _)) = ns
comInWeightSum :: Community c -> InWeightSum
comInWeightSum (Community (_, inWeightSum, _, _)) = inWeightSum
comTotWeightSum :: Community c -> TotWeightSum
comTotWeightSum (Community (_, _, totWeightSum, _)) = totWeightSum
comLabel :: Community c -> c
comLabel (Community (_, _, _, c)) = c

type CGrNode = Node
type CGrEdge = (InWeightSum, TotWeightSum)

type CGr a = Gr (Community a) ()

graphWeight :: FGraph a b -> GraphWeightSum
graphWeight gr = GraphWeightSum $ 0.5 * ufold (\(_, n, _, _) -> weight' $ context gr n) 0 gr
  where
    weight' (p, _, _, s) acc = acc + (sumEdgeWeights $ p <> s)
